#ifndef METRO_GRAPH_H_INCLUDED
#define METRO_GRAPH_H_INCLUDED

#include <boost/graph/graph_traits.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/dijkstra_shortest_paths.hpp>
#include <boost/format.hpp>

#include <iostream>
#include <fstream>

namespace metro
{

/**
 * \class Graph
 * \brief Représente le graph du plan de metro
 */
template<typename sommet>
class Graph
{
	typedef boost::adjacency_list< boost::listS
	                             , boost::vecS
						         , boost::undirectedS
						         , boost::no_property
								 , boost::property<boost::edge_weight_t, int>
						         > graph_t;

	typedef boost::graph_traits<graph_t>::vertex_descriptor vertex_descriptor;
	
	typedef boost::graph_traits<graph_t>::edge_descriptor edge_descriptor;
	
	typedef std::pair<sommet, sommet> edge_t;

	typedef std::list<edge_t> list_edge_t;
	typedef std::list<int>    list_weight_t;
	
	list_edge_t    m_edges;
	list_weight_t  m_weigths;
	size_t         m_nbSommets;

public:
	typedef std::list< std::pair<sommet, size_t> > solution_t;
	
	Graph()
	: m_nbSommets(0)
	{
	}
	
	~Graph()
	{
	}
	
	/**
	 * \param s1     1er sommet
	 * \param s2     2nd sommet
	 * \param weight distance entre les 2 sommets
	 */
	void add_edge(sommet s1, sommet s2, int weight)
	{
		m_edges.push_back(edge_t(s1, s2));
		m_weigths.push_back(weight);
	}
	
	void set_nb_sommets(size_t nbSommets)
	{
		m_nbSommets = nbSommets;
	}
	
	bool find_way(sommet s1, sommet s2, solution_t& trajet)
	{
		assert(m_edges.size() == m_weigths.size());

		// Creation du graph
		graph_t g(m_edges.begin(), m_edges.end(), m_weigths.begin(), m_nbSommets);
		boost::property_map<graph_t, boost::edge_weight_t>::type weightmap = boost::get(boost::edge_weight, g);

		// On initialise la recherche a partir de s1 (sommet initial)
		std::vector<vertex_descriptor> precedents(num_vertices(g));
		std::vector<int> distances(num_vertices(g));
		vertex_descriptor s = vertex(s1, g);

		// Calcul des chemins les plus court
		boost::dijkstra_shortest_paths( g
									  , s
									  , boost::predecessor_map(&precedents[0]).distance_map(&distances[0])
									  );

		// Solution
		trajet.push_front(std::make_pair(s2, distances[s2]));
        sommet cur  = s2;
        sommet prec = 0;
        do {
			prec = precedents[cur];
			if (prec == cur) {
				return false;
			}
			else if (prec != s1) {
				trajet.push_front(std::make_pair(prec, distances[prec]));
				cur = prec;
			}
        }
        while (prec != s1);
		return true;
	}
	
};

}

#endif // METRO_GRAPH_H_INCLUDED
