#ifndef METRO_STATION_DATA_H_INCLUDED
#define METRO_STATION_DATA_H_INCLUDED

#include <sstream>
#include <boost/multi_index_container.hpp>
#include <boost/multi_index/member.hpp>
#include <boost/multi_index/mem_fun.hpp>
#include <boost/multi_index/ordered_index.hpp>

namespace metro
{

namespace bmi = boost::multi_index;

/// ID d'une station
typedef unsigned int id_station;

namespace station
{

/**
 * \class Data
 */
class Data
{
	friend class IdMgr;
	
	std::string   m_ligne;
	std::string   m_nom;
	int           m_nEtape;
	id_station    m_id;

public:
	/// Identifiant d'un ID invalid 
	static const id_station invalid_id;

	Data(const std::string& ligne, const std::string& nom, int nEtape);

	/// Retourne la clef unique d'une station
	std::string key() const;

	/// Retourn l'id d'une station
	id_station id() const;

	/// Retourne le nom de la station
	const std::string& nom() const;

	/// Retourne le nom de la ligne à laquelle appartient la station
	const std::string& ligne() const;

	/// Retourne le numéro d'étape de la station
	int etape() const;

	/// operateur de comparaison, permettant le stockage dans une map
	bool operator<(const Data& other);
};

/// Permet l'indexation du container de station par critere d'ID
struct id  {};
/// Permet l'indexation du container de station par critere de clef
struct key {};

/// Container de station, indexable par plusieurs criteres (id et clef)
typedef boost::multi_index_container<
	Data,
	bmi::indexed_by<
		bmi::ordered_unique<
			bmi::tag<id>, BOOST_MULTI_INDEX_CONST_MEM_FUN(Data, id_station, id)
		>,
		bmi::ordered_non_unique<
			bmi::tag<key>, BOOST_MULTI_INDEX_CONST_MEM_FUN(Data, std::string, key)
		>
	>
> IdContainer;

/**
 * \class IdMgr
 * 
 * Cette classe permet de gerer les ID des stations
 */
struct IdMgr
{
	static IdContainer m_idStations;

	/// Recupere (ou genere si il n'existe pas) un ID unique pour une station
	static id_station get_id(Data& stationData);

	/// Retourne le container de station indexé par ID
	static bmi::index<IdContainer, id>::type& get_by_id();
};

/**
 * Raccourci pour station::IdMgr::get_id()
 */
id_station get_id(Data& stationData);

typedef bmi::index<IdContainer, id>::type container_by_id_t;
typedef IdContainer::index<id>::type set_by_id_t;

/**
 * Raccourci pour station::IdMgr::get_by_id()
 */
const container_by_id_t& get_by_id();


} // namespace metro::station

/**
 * \class StationData
 */
class StationData : public station::Data
{
public:
	StationData(const std::string& ligne, const std::string& nom, int nEtape)
	: station::Data(ligne, nom, nEtape)
	{
		station::get_id(*this);
	}
};

} // namespace metro

#endif // METRO_STATION_DATA_H_INCLUDED
