#include "metro/plan.h"
#include <iostream>
#include <fstream>
#include <boost/foreach.hpp>
#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/trim.hpp>
#include <boost/algorithm/string/join.hpp>
#include <boost/algorithm/string/predicate.hpp>
#include <boost/algorithm/string/case_conv.hpp>
#include <boost/assign/list_of.hpp>

using namespace metro;

/**
 * \param delaiStation delai entre 2 stations en seconde
 * \param delaiCorrespondance delai entre 2 correspondance en seconde
 * \param bOptimise optimise la recherche
 */
Plan::Plan(size_t delaiStation, size_t delaiCorrespondance, bool bOptimise)
: m_delaiStation(delaiStation)
, m_delaiCorrespondance(delaiCorrespondance)
, m_bOptimise(bOptimise)
{
}

/**
 * Les etapes sont les stations par lesquelles le trajet doit obligatoirement
 * passer. Si l'option 'optimise' est présente, l'ordre des étapes n'est pas 
 * forcement respecté.
 * 
 * \param etapes
 * 
 * \return true si les etapes existent dans le plan
 */
bool Plan::SetEtapes(const etapes_t& etapes)
{
	BOOST_FOREACH(const std::string& sEtape, etapes) {
		std::string etape = Clean(sEtape);
		if (StationExists(etape) == false) {
			std::cerr << "l'etape " << sEtape << " n'existe pas " << std::endl;
			return false;
		}
	}
	m_etapes = etapes;
	return true;
}

/**
 * Le fichier contient l'ensemble des stations rapportées à une ligne
 * donnée. Ce fichier est codé en UTF-16 avec BOM
 * 
 * \param filename nom du fichier
 * 
 * \return true si le fichier est correctement chargé et valide
 */
bool Plan::Load(const std::string& filename)
{
	std::ifstream iFile(filename.c_str(), std::ios_base::binary);
	if(!iFile) {
		std::cerr << "Impossible d'ouvrir le fichier " << filename << std::endl;
		return false;
	}
	
	char buffer[2] = { 0 };
	// Verification du BOM
	iFile.read(buffer, 2);
	if (iFile.eof() || buffer[0] != -1 || buffer[1] != -2) {
		std::cerr << "Le fichier " << filename << " est invalide (probleme de BOM)" << std::endl;
		return false;
	}

	// Lecture du contenu du fichier
	std::string content;
	while (true) {
		iFile.read(buffer, 2);
		if (iFile.eof()) {
			break;
		}
		if (buffer[1]) {
			std::cerr << "Le fichier " << filename << " est invalide (UTF-16 invalide)" << std::endl;
			return false;
		}
		content += buffer[0]; // C'est pas beau, mais ca marche
	}
	
	// Chaque ligne devient un element d'un vecteur
	std::vector<std::string> fileContent;
	boost::split(fileContent, content, boost::is_any_of("\n\r"));
	
	// Analyse ligne / station
	std::string curLigne;
	BOOST_FOREACH(std::string& fileLigne, fileContent) {
		boost::trim(fileLigne);
		if (fileLigne[0] == '[') {
			// Definition d'une ligne
			if (*fileLigne.rbegin() != ']') {
				std::cerr << "Le fichier " << filename << " est invalide (section non terminee)" << std::endl;
				return false;
			}
			curLigne = fileLigne.substr(1, fileLigne.size() - 2);
		}
		else if (!fileLigne.empty()) {
			// C'est une station
			std::string station = Clean(fileLigne);
			m_fileNames[std::make_pair(curLigne, station)] = fileLigne;
			Add(curLigne, station);
		}
	}
	return true;
}

/**
 * Cette fonction permet de remplacer tous les caracteres speciaux d'une
 * station par un caractere simple. Ceci permet de pouvoir nommer une
 * station avec ou sans accent / tirets / majuscules
 * 
 * \param station nom de la station. Cette variable est modifiée lors de 
 * l'appel à la fonction
 */
std::string Plan::Clean(const std::string& station)
{
	typedef std::map<char,char> MapConversion;
	static const MapConversion conversion = boost::assign::map_list_of
		('-' , ' ')  // On supprime les tirets 
		(0xC8, 'e')  // È
		(0xC9, 'e')  // É
		(0xE0, 'a')  // à
		(0xE1, 'a')  // á
		(0xE2, 'a')  // â
		(0xE7, 'c')  // ç
		(0xE8, 'e')  // è
		(0xE9, 'e')  // é
		(0xEA, 'e')  // ê
		(0xEB, 'e')  // ë
		(0xEE, 'i')  // î
		(0xF4, 'o')  // ô
		(0xF9, 'u')  // ù
		(0xFB, 'u')  // û
		(0xFC, 'u')  // ü
		;

	std::string cleaned(station);
	BOOST_FOREACH(char& c, cleaned) {
		MapConversion::const_iterator itConv = conversion.find(c);
		if (itConv != conversion.end()) {
			c = itConv->second;
		}
	}
	return boost::to_lower_copy(cleaned);
}

/**
 * \param temps temps en seconde
 * 
 * \return une chaine de caractere au format HH:MM:SS
 */
std::string Plan::FormatTemps(size_t temps)
{
	return (boost::format( "%02d:%02d:%02d")
	                     % (temps / 3600)
						 % ((temps / 60) % 60)
						 % (temps % 60)
						 ).str();
}

/**
 * \param station nom de la station
 * 
 * \return true si la station existe su le reseau
 */
bool Plan::StationExists(const std::string& station) const
{
	return m_correspondances.find(station) != m_correspondances.end();
}

/**
 * \param stationData information sur la station
 * 
 * \return true si la station est une etape
 */
bool Plan::IsStationEtape(const station::Data& stationData) const
{
	return m_etapesId.find(stationData.id()) != m_etapesId.end();
}

/**
 * \param ligne nom de la ligne
 * \param station nom de la station
 * 
 * \return le nom de la station tel qu'il apparait dans le fichier
 */
const std::string& Plan::GetFileStationName(const std::string& ligne, const std::string& station) const
{
	file_names_t::const_iterator itNom = m_fileNames.find(std::make_pair(ligne, station));
	if (itNom != m_fileNames.end())
		return itNom->second;
	return station;
}

/**
 * \param g           graph
 * \param stationData information sur la station
 * \param temps       durée en seconde entre la station et les correspondances
 * \param nEtape      numéro d'etapes à laquelle on veux se connecter
 */ 
void Plan::CreateCorrespondance(graph_t& g, const StationData& stationData, int temps, int nEtape)
{
	const std::string& station = stationData.nom();
	correspondances_t::const_iterator itStation = m_correspondances.lower_bound(station);
	for ( ; itStation != m_correspondances.upper_bound(station); ++itStation) {
		StationData correspondance(itStation->second, station, nEtape);
		if (correspondance.id() != stationData.id()) {
			g.add_edge(stationData.id(), correspondance.id(), temps);
		}
	}
}

/**
 * Ajoute une station de metro sur une ligne donnée. L'ordre est important, car 
 * la station est ajoutée à la fin des stations existantes.
 * 
 * \param ligne numéro de la ligne de métro
 * \param station nom de la station de métro
 */
void Plan::Add(const std::string& ligne, const std::string& station)
{
	m_plan[ligne].push_back(station);
	m_correspondances.insert(correspondances_t::value_type(station, ligne));
}

/**
 * Cela crée les différent arcs du graph pour les lignes de metro, et pour les
 * correspondances
 * 
 * \param g      graph
 * \param nEtape numéro d'étape
 * \param etapes etapes restant à traiter
 */
void Plan::Create(graph_t& g, int& nEtape, const etapes_t& etapes)
{
	BOOST_FOREACH(plan_t::value_type pairLigneStations, m_plan) {
		const std::string& nomLigne  = pairLigneStations.first;
		const ligne_t& stations      = pairLigneStations.second;

		for (ligne_t::const_iterator itStation = stations.begin(); itStation != stations.end(); ++itStation) {
			// Creation de la station
			StationData stationData(nomLigne, *itStation, nEtape);
			ligne_t::const_iterator itStation2 = itStation;
			if (++itStation2 != stations.end()) {
				StationData stationData2(nomLigne, *itStation2, nEtape);
				// On crée le chemin entre 2 stations successives sur une ligne
				g.add_edge(stationData.id(), stationData2.id(), m_delaiStation);
			}
			// Creation des correspondances
			CreateCorrespondance(g, stationData, m_delaiCorrespondance, nEtape);
		}
	}
	
	if (etapes.empty()) {
		m_etapesFinales.push_back(nEtape);
		return;
	}
	
	// On ajoute les étapes
	if (m_bOptimise) {
		// Parcours optimisé, on crée un graph par etape
		int old_etape = nEtape;
		BOOST_FOREACH(const std::string& sEtape, etapes) {
			nEtape++;
			int new_etape = nEtape;
			// Création de la station etape
			StationData etapeData(boost::str(boost::format("_ETAPE_ %1%") % nEtape), sEtape, nEtape);
			m_etapesId.insert(etapeData.id());
			// Creation des correspondances entre l'ancien graph et l'etape
			CreateCorrespondance(g, etapeData, 0, old_etape);
			// Creation du nouveau graph (plan de metro) pour l'etape suivante
			etapes_t etapes_restantes; // TODO : optimiser avec remove_if (?)
			BOOST_FOREACH(const std::string& sEtape2, etapes) {
				if (sEtape != sEtape2) {
					etapes_restantes.push_back(sEtape2);
				}
			}
			Create(g, nEtape, etapes_restantes);
			// Creation des correspondances entre le nouveau graph et l'etape
			CreateCorrespondance(g, etapeData, m_delaiCorrespondance, new_etape);
		}
	}
	else {
		// Non optimisé : on parcours les étapes dans l'ordre
		int old_etape = nEtape;
		nEtape++;
		// Création de la station etape
		StationData etapeData(boost::str(boost::format("_ETAPE_ %1%") % nEtape), etapes[0], nEtape);
		m_etapesId.insert(etapeData.id());
		// Creation des correspondances entre l'ancien graph et l'etape
		CreateCorrespondance(g, etapeData, 0, old_etape);
		// Creation du nouveau graph (plan de metro) pour l'etape suivante
		etapes_t etapes_restantes(++etapes.begin(), etapes.end());
		Create(g, nEtape, etapes_restantes);
		// Creation des correspondances entre le nouveau graph et l'etape
		CreateCorrespondance(g, etapeData, m_delaiCorrespondance, old_etape+1);
	}
}

/**
 * effectue la recherche du trajet le plus court entre 2 stations. L'algorithme
 * utilisé celui de Dijkstra.
 * 
 * \param sDepart   nom de la station de départ
 * \param sArrivee  nom de la station d'arrivée
 */
void Plan::Trajet(const std::string& sDepart, const std::string& sArrivee)
{
	std::string depart  = Clean(sDepart);
	std::string arrivee = Clean(sArrivee);
	
	// Verifications que les stations existent et sont differentes
	if (StationExists(depart) == false) {
		std::cerr << sDepart << " n'est pas une station de metro" << std::endl;
		return;
	}
	
	if (StationExists(arrivee) == false) {
		std::cerr << sArrivee << " n'est pas une station de metro" << std::endl;
		return;
	}
	
	if (depart == arrivee) {
		std::cout << "Ne cherchez pas, vous y etes !" << std::endl;
		return;
	}
	
	std::cout << "Trajet de " << depart << " vers " << arrivee << std::endl;
	if (!m_etapes.empty()) {
		std::cout << "Etapes : " << boost::join(m_etapes, ", ") << std::endl;
	}
	if (m_bOptimise) {
		std::cout << "Recherche optimisee" << std::endl;
	}
	std::cout << std::endl;
	
	// On créé le graph
	graph_t graph;
	
	// On crée un plan initial
	int nEtape = 0;
	Create(graph, nEtape, m_etapes);
	
	// Creation de 2 nouvelles station virtuelles : depart et arrivée
	StationData departData("_DEPART_", depart, 0);
	CreateCorrespondance(graph, departData, 0, 0);
	StationData arriveeData("_ARRIVEE_", arrivee, nEtape + 1);
	
	// On relie l'arrivée à l'ensemble des graphes terminaux
	BOOST_FOREACH(int nEtapeFinale, m_etapesFinales) {
		CreateCorrespondance(graph, arriveeData, 0, nEtapeFinale);
	}

	// Le nombre de sommet est le dernier id créé + 1
	graph.set_nb_sommets(arriveeData.id() + 1);
	
	// Algo de recherche du trajet le plus court
	solution_t trajet;
	if (graph.find_way(departData.id(), arriveeData.id(), trajet) == false) {
		std::cout << "Aucun moyen d'effectuer le trajet" << std::endl;
		return;
	}
	
	const station::container_by_id_t& c = station::get_by_id();

	// Suppression des stations inutiles, ie : d'une meme ligne
	solution_t::iterator itIdStationRemovable = trajet.begin();
	std::string currentLigne;
	int         currentEtape = 0;
	do {
		station::set_by_id_t::const_iterator itStationData = c.find(itIdStationRemovable->first);
		assert(itStationData != c.end());
		if (itStationData->ligne() == currentLigne && itStationData->etape() == currentEtape) {
			// On est toujours sur la meme ligne : on supprime cette station
			itIdStationRemovable = trajet.erase(itIdStationRemovable);
		}
		else {
			// On a changé de ligne
			currentLigne = itStationData->ligne();
			currentEtape = itStationData->etape();
			++itIdStationRemovable;
		}
	}
	while (itIdStationRemovable != trajet.end());

	// Affichage de la solution
	for (solution_t::const_iterator itIdStation = trajet.begin(); itIdStation->first != arriveeData.id(); ++itIdStation) {
		std::cout << FormatTemps(itIdStation->second) << " : ";
		station::set_by_id_t::const_iterator itStationData = c.find(itIdStation->first);
		if (IsStationEtape(*itStationData)) {
			// nous sommes a une etape
			std::cout << "Arrivee a l'etape " 
			          << GetFileStationName(itStationData->ligne(), itStationData->nom())
					  << std::endl;
			continue;
		}
		
		std::cout << "Prendre la ligne [" << itStationData->ligne() << "] direction ";
		// Quelle est la station ou l'on doit changer : la suivante
		solution_t::const_iterator itIdStationChangement = itIdStation;
		++itIdStationChangement;
		station::set_by_id_t::const_iterator itStationChangementData = c.find(itIdStationChangement->first);
		// Recherche de la direction
		const ligne_t& curLigne = m_plan[itStationData->ligne()];
		ligne_t::const_iterator itStation = std::find(curLigne.begin(), curLigne.end(), itStationData->nom());
		if (std::find(itStation, curLigne.end(), itStationChangementData->nom()) == curLigne.end()) {
			// Le changement de ligne n'est pas à la suite de la station en cours, la 
			// direction est donc le début de la ligne
			std::cout << GetFileStationName(itStationData->ligne(), curLigne.front());
		}
		else {
			std::cout << GetFileStationName(itStationData->ligne(), curLigne.back());
		}
		std::cout << std::endl;
		std::cout << "  Sortir a la station " 
		          << GetFileStationName(itStationChangementData->ligne(),itStationChangementData->nom())
				  << std::endl;
	}
	std::cout << FormatTemps(trajet.rbegin()->second) << " : Arrivee" << std::endl;
}
